from copy import deepcopy

from django.apps import apps
from django.contrib.contenttypes.models import ContentType


class ContentTypeCacheFixMixin:
    """
    TestCase mixin to fill the ContentType cache to avoid flaky database queries.

    The Problem is, that the ContentTypeManager used a shared cache for lookups.
    If your tests checks the made database queries, then you may get different results
    depending on how the cache is filled.

    With this Mixin the ContentTypeManager cache is always filled and all tests will never
    hit the databases...
    """
    @classmethod
    def _fill_content_type_cache(cls):
        # Fill ContentTypeManager cache
        app_configs = apps.get_app_configs()
        for app_config in app_configs:
            models = app_config.get_models()
            for model in models:
                ContentType.objects.get_for_model(model)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._fill_content_type_cache()
        cls._content_type_cache = deepcopy(ContentType.objects._cache)

    def setUp(self):
        super().setUp()
        ContentType.objects._cache = deepcopy(self._content_type_cache)



import logging
import zoneinfo

from django.test import RequestFactory, SimpleTestCase
from django.utils import timezone

from bx_django_utils.user_timezone.middleware import UserTimezoneMiddleware


DEFAULT_TIMEZONE = zoneinfo.ZoneInfo('America/Los_Angeles')


class UserTimezoneMiddlewareTestCase(SimpleTestCase):
    @timezone.override(DEFAULT_TIMEZONE)
    def test_middleware(self):
        self.assertEqual(timezone.get_current_timezone(), DEFAULT_TIMEZONE)

        def fake_get_response(request):
            return f'User timezone: "{timezone.get_current_timezone()}"'

        middleware = UserTimezoneMiddleware(get_response=fake_get_response)
        fake_request = RequestFactory().get('/foo/bar')

        ##########################################################################
        # No cookie set -> don't change current/default timezone:
        fake_response = middleware(request=fake_request)
        self.assertEqual(fake_response, 'User timezone: "America/Los_Angeles"')
        self.assertEqual(timezone.get_current_timezone(), DEFAULT_TIMEZONE)  # not changed?

        ##########################################################################
        # Temporary change timezone via cookie to: 'Europe/Berlin'
        fake_request.COOKIES['UserTimeZone'] = 'Europe/Berlin'
        fake_response = middleware(request=fake_request)
        self.assertEqual(fake_response, 'User timezone: "Europe/Berlin"')
        self.assertEqual(timezone.get_current_timezone(), DEFAULT_TIMEZONE)  # reset?

        ##########################################################################
        # Temporary change timezone via cookie to: 'Europe/London'
        fake_request.COOKIES['UserTimeZone'] = 'Europe/London'
        fake_response = middleware(request=fake_request)
        self.assertEqual(fake_response, 'User timezone: "Europe/London"')
        self.assertEqual(timezone.get_current_timezone(), DEFAULT_TIMEZONE)  # reset?

        ##########################################################################
        # If nonsense is in cookie -> just log a error and don't change the default timezone:
        fake_request.COOKIES['UserTimeZone'] = 'non/sense'
        with self.assertLogs(logger='bx_django_utils', level=logging.ERROR) as logs:
            fake_response = middleware(request=fake_request)

        # Default time zone?
        self.assertEqual(fake_response, 'User timezone: "America/Los_Angeles"')
        self.assertEqual(timezone.get_current_timezone(), DEFAULT_TIMEZONE)

        self.assertEqual(
            logs.output,
            [
                "ERROR:bx_django_utils.user_timezone.middleware:"
                "Unknown user time zone 'non/sense'"
            ],
        )



from urllib.parse import quote_plus, unquote_plus

from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Count


class NotAllSimpleListFilter(admin.SimpleListFilter):
    """
    Similar to SimpleListFilter, but don't add "All" choice.
    """

    def choices(self, changelist):
        """
        The origin SimpleListFilter.choices() inserts "All" as first choice.
        But we would like to have other filter as default.
        """
        for lookup, title in self.lookup_choices:
            yield {
                'selected': self.value() == lookup,
                'query_string': changelist.get_query_string({self.parameter_name: lookup}),
                'display': title,
            }


class ExistingCountedListFilter(admin.SimpleListFilter):
    """
    Advanced SimpleListFilter that list only existing filter values with counts.
    """

    model_field_name = None  # Model field name to filter by
    value_fmt = '{value} ({count})'  # Used to generate the filter entry line

    def __init__(self, request, params, model, model_admin):
        if self.model_field_name is None:
            raise ImproperlyConfigured(
                f"List filter '{self.__class__.__name__}' does not specify a 'model_field_name'."
            )
        super().__init__(request, params, model, model_admin)

    def lookups(self, request, model_admin: admin.ModelAdmin):
        qs = model_admin.get_queryset(request)
        qs = (
            qs.order_by(self.model_field_name)
            .values(self.model_field_name)
            .annotate(count=Count('pk'))
        )
        return [
            (
                quote_plus(filter_entry[self.model_field_name]),
                self.value_fmt.format(
                    value=filter_entry[self.model_field_name], count=filter_entry['count']
                ),
            )
            for filter_entry in qs
        ]

    def queryset(self, request, queryset):
        if filter_value := self.value():
            filter_value = unquote_plus(filter_value)
            filter_kwargs = {self.model_field_name: filter_value}
            return queryset.filter(**filter_kwargs)



import io
import logging

from django.contrib.admin import AdminSite
from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import RequestFactory, SimpleTestCase, TestCase
from django.urls import reverse
from django.views import View

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.checks import admin_extra_views_check
from bx_django_utils.admin_extra_views.conditions import only_staff_user
from bx_django_utils.admin_extra_views.datatypes import _APP_LABELS, _URL_NAMES, AdminExtraMeta, PseudoApp
from bx_django_utils.admin_extra_views.management.commands import admin_extra_views
from bx_django_utils.admin_extra_views.registry import extra_view_registry, register_admin_view
from bx_django_utils.admin_extra_views.site import ExtraViewAdminSite
from bx_django_utils.admin_extra_views.utils import reverse_admin_extra_view
from bx_django_utils.admin_extra_views.views import Redirect2AdminExtraView
from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin, assert_html_response_snapshot
from bx_django_utils.test_utils.users import make_max_test_user, make_minimal_test_user
from bx_django_utils_tests.test_app.admin_views import (
    DemoView1,
    DemoView2,
    DemoView3,
    only_john_can_access,
    pseudo_app1,
    pseudo_app2,
)


class ClearDataTypesInfoMixin:
    def reset(self):
        _APP_LABELS.clear()
        _URL_NAMES.clear()

    def setUp(self) -> None:
        super().setUp()
        self.reset()

    def tearDown(self) -> None:
        super().tearDown()
        self.reset()


class DataTypesTestCase(ClearDataTypesInfoMixin, SimpleTestCase):

    def test_pseudo_app(self):
        PseudoApp(meta=AdminExtraMeta(name='Pseudo App'))

        with self.assertRaisesMessage(
            AssertionError,
            "PseudoApp must be have a unique label! Current label is: 'pseudo-app'",
        ):
            PseudoApp(meta=AdminExtraMeta(name='Pseudo App'))

    def test_unique_urls(self):
        pseudo_app = PseudoApp(meta=AdminExtraMeta(name='Pseudo App'))

        AdminExtraMeta(name='Pseudo Model').setup_app(pseudo_app)

        with self.assertRaisesMessage(
            AssertionError,
            "App label combination 'pseudo-app_pseudo-model' is not unique!",
        ):
            AdminExtraMeta(name='Pseudo Model').setup_app(pseudo_app)


class AdminTestCase(HtmlAssertionMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        # Has all permissions but is not staff member:
        cls.non_staff_user = make_max_test_user(is_staff=False, exclude_permissions=())

        cls.staff_user = make_minimal_test_user(username='foobar', is_staff=True, permissions=())

        cls.superuser = User.objects.create_superuser(
            username='a-superuser', password='ThisIsNotAPassword!'
        )

    def test_permissions(self):
        # Default condition added?
        self.assertEqual(pseudo_app1.meta.conditions, {only_staff_user})
        self.assertEqual(DemoView1.meta.conditions, {only_staff_user})

        self.assertEqual(DemoView1.meta.url_name, 'pseudo-app-1_demo-view-1')
        url = reverse('pseudo-app-1_demo-view-1')
        self.assertEqual(url, '/admin/pseudo-app-1/demo-view-1/')

        # Anonymous user can't access:
        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs, self.assertLogs(
            'django.request', level=logging.WARNING
        ):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        assert logs.output == [
            "WARNING:bx_django_utils.admin_extra_views.base_view:"
            "User (pk:None) did not pass 'only_staff_user' for 'DemoView1'"
        ]

        # Non-staff users can also not access:
        self.client.force_login(self.non_staff_user)
        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs, self.assertLogs(
            'django.request', level=logging.WARNING
        ):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        assert logs.output == [
            "WARNING:bx_django_utils.admin_extra_views.base_view:"
            f"User (pk:{self.non_staff_user.pk}) did not pass 'only_staff_user' for 'DemoView1'"
        ]

        # Staff user can use the view:
        self.client.force_login(self.staff_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assert_html_parts(
            response,
            parts=('<p>Just the demo view 1</p>',),
        )

    def test_view_inherits_app_conditions(self):
        # App will have default conditions:
        self.assertEqual(pseudo_app2.meta.conditions, {only_staff_user})

        # The app conditions will be added to the view conditions:
        self.assertEqual(DemoView3.meta.conditions, {only_staff_user, only_john_can_access})

    def test_index_page(self):
        app1_demo1 = '<a href="/admin/pseudo-app-1/demo-view-1/">Demo View 1</a>'
        app1_demo2 = '<a href="/admin/pseudo-app-1/demo-view-2/">Demo View 2</a>'
        app2_demo3 = '<a href="/admin/pseudo-app-2/demo-view-3/">Demo View 3</a>'

        ############################################################################
        # A super user sees all pseudo apps/models:
        self.client.force_login(self.superuser)
        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                '<title>Site administration | Django site admin</title>',
                (
                    '<a href="" class="section"'
                    ' title="Models in the Pseudo App 2 application">Pseudo App 2</a>'
                ),
                app1_demo1,
                app1_demo2,
                app2_demo3,
            ),
        )
        assert_html_response_snapshot(response, query_selector='#content-main', validate=False)

        ############################################################################
        # Staff user that is not "john" can not see/access demo view 3:
        self.client.force_login(self.staff_user)
        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                app1_demo1,
                app1_demo2,
            ),
        )
        self.assert_parts_not_in_html(response, parts=(app2_demo3,))
        with self.assertLogs('bx_django_utils', level=logging.WARNING) as logs, self.assertLogs(
            'django.request', level=logging.WARNING
        ):
            response = self.client.get('/admin/pseudo-app-2/demo-view-3/')
        self.assertEqual(response.status_code, 403)
        assert logs.output == [
            "WARNING:bx_django_utils.admin_extra_views.base_view:"
            f"User (pk:{self.staff_user.pk}) did not pass 'only_john_can_access' for 'DemoView3'"
        ]

        ############################################################################
        # Only "john" can see/access demo view 3:
        self.client.force_login(
            make_minimal_test_user(username='john', is_staff=True, permissions=())
        )
        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                app1_demo1,
                app1_demo2,
                app2_demo3,
            ),
        )
        response = self.client.get('/admin/pseudo-app-2/demo-view-3/')
        self.assertEqual(response.status_code, 200)
        self.assert_html_parts(
            response,
            parts=("Just the demo view at '/admin/pseudo-app-2/demo-view-3/'",),
        )


class AdminExtraViewSimpleTestCase(SimpleTestCase):
    def test_reverse_admin_extra_view(self):
        url = reverse_admin_extra_view(DemoView2)
        self.assertEqual(url, '/admin/pseudo-app-1/demo-view-2/')

    def test_redirect_view(self):
        redirect_view = Redirect2AdminExtraView.as_view(admin_view=DemoView2)

        dummy_request = RequestFactory().request()
        response = redirect_view(request=dummy_request)
        self.assertRedirects(
            response, expected_url='/admin/pseudo-app-1/demo-view-2/', fetch_redirect_response=False
        )

    def test_redirect_view_via_urls(self):
        response = self.client.get('/old_demo_3/')
        self.assertRedirects(
            response, expected_url='/admin/pseudo-app-2/demo-view-3/', fetch_redirect_response=False
        )


class ManageCommandTestCase(TestCase):
    def test_admin_extra_views_command(self):
        capture_stdout = io.StringIO()
        capture_stderr = io.StringIO()
        call_command(admin_extra_views.Command(), stdout=capture_stdout, stderr=capture_stderr)
        stdout_output = capture_stdout.getvalue()
        stderr_output = capture_stderr.getvalue()

        self.assertIn('Pseudo App 1', stdout_output)
        self.assertIn('Demo View 1', stdout_output)
        self.assertIn('/admin/pseudo-app-1/demo-view-1/', stdout_output)
        self.assertEqual(stderr_output, '')


class ChecksTestCase(ClearDataTypesInfoMixin, TestCase):
    def setUp(self):
        self._old_pseudo_apps = extra_view_registry.pseudo_apps.copy()
        super().setUp()

    def tearDown(self):
        super().tearDown()
        extra_view_registry.pseudo_apps = self._old_pseudo_apps

    def test_no_errors(self):
        errors = admin_extra_views_check(app_configs=None)
        self.assertEqual(errors, [])

    def test_custom_admin_site(self):
        class CustomAdminSite(ExtraViewAdminSite):
            pass

        @register_admin_view(
            pseudo_app=PseudoApp(
                meta=AdminExtraMeta(name='Foo'),
                admin_site=CustomAdminSite(),
            )
        )
        class TestView(AdminExtraViewMixin, View):
            meta = AdminExtraMeta(name='Bar')

        errors = admin_extra_views_check(app_configs=None)

        # We should only get the "URL reverse error" because CustomAdminSite is not in urls.py ;)
        self.assertEqual(len(errors), 1)
        reverse_error = errors[0]
        self.assertEqual(reverse_error.id, 'admin_extra_views.E001')

    def test_wrong_admin_site(self):
        @register_admin_view(
            pseudo_app=PseudoApp(
                meta=AdminExtraMeta(name='Foo'),
                admin_site=AdminSite(),
            )
        )
        class TestView(AdminExtraViewMixin, View):
            meta = AdminExtraMeta(name='Bar')

        errors = admin_extra_views_check(app_configs=None)
        self.assertEqual(len(errors), 2)
        reverse_error = errors[0]
        self.assertEqual(reverse_error.id, 'admin_extra_views.E001')
        self.assertEqual(reverse_error.msg, "Admin extra views URL reverse error with 'TestView'")

        admin_class_error = errors[1]
        self.assertEqual(admin_class_error.id, 'admin_extra_views.E002')
        self.assertEqual(
            admin_class_error.msg,
            "Pseudo app 'Foo' error: Admin site 'AdminSite'"
            " is not a instance of ExtraViewAdminSite!",
        )



"""
    Just a DEMO for bx_django_utils.admin_extra_views
"""
from django.contrib import messages
from django.contrib.auth.models import User
from django.http import HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import TemplateView, View

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta, PseudoApp
from bx_django_utils.admin_extra_views.registry import register_admin_view
from bx_django_utils.feature_flags.admin_views import ManageFeatureFlagsBaseView
from bx_django_utils_tests.test_app.feature_flags import bar_feature_flag, foo_feature_flag


def only_john_can_access(request):
    user: User = request.user
    if user.is_superuser:
        return True
    if user.username == 'john':
        return True
    else:
        return False


pseudo_app1 = PseudoApp(meta=AdminExtraMeta(name='Pseudo App 1'))
pseudo_app2 = PseudoApp(meta=AdminExtraMeta(name='Pseudo App 2'))


@register_admin_view(pseudo_app=pseudo_app1)
class DemoView1(AdminExtraViewMixin, TemplateView):
    meta = AdminExtraMeta(name='Demo View 1')
    template_name = 'admin_extra_views/demo1.html'

    def get_context_data(self, **context):
        context = super().get_context_data(**context)
        context['content'] = 'Just the demo view 1'
        return context


@register_admin_view(pseudo_app=pseudo_app1)
class DemoView2(AdminExtraViewMixin, View):
    meta = AdminExtraMeta(name='Demo View 2')

    def get(self, request):
        return HttpResponse('Just the demo view 2')


@register_admin_view(pseudo_app=pseudo_app2)
class DemoView3(AdminExtraViewMixin, View):
    meta = AdminExtraMeta(
        name='Demo View 3',
        conditions={only_john_can_access},  # Note: pseudo_app2 conditions will be added!
    )

    def get(self, request):
        url = reverse(self.meta.url_name)
        return HttpResponse(f'Just the demo view at {url!r}')


#####################################################################################
# Feature Flags DEMO

manage_flags_app = PseudoApp(
    meta=AdminExtraMeta(
        name=_('Feature Flags'),
        app_label='feature_flags',
        # Set your needed conditions here!
    )
)


@register_admin_view(pseudo_app=manage_flags_app)
class ManageFeatureFlagsAdminExtraView(ManageFeatureFlagsBaseView):
    pass


@register_admin_view(pseudo_app=manage_flags_app)
class FeatureFlagsInfoView(AdminExtraViewMixin, View):
    """
    Just display the states of the demo feature flags
    """

    meta = AdminExtraMeta(name='Feature Flags values DEMO')

    def get(self, request):
        messages.info(self.request, f'Demo Feature "Foo" state: {foo_feature_flag.state.name}')
        messages.info(self.request, f'Demo Feature "Bar" state: {bar_feature_flag.state.name}')
        url = reverse('admin:index')
        return HttpResponseRedirect(url)



from django.db.models import CharField
from django.utils.translation import gettext_lazy as _
from stdnum import ean

from bx_django_utils.data_types.gtin.constants import DEFAULT_ACCEPT_LENGTH
from bx_django_utils.data_types.gtin.form_fields import GtinFormField
from bx_django_utils.data_types.gtin.validators import GtinValidator


class GtinModelField(CharField):
    """
    GTIN model field
    """
    description = _('GTIN number')

    def __init__(self, *args, accepted_length=DEFAULT_ACCEPT_LENGTH, **kwargs):
        self.accepted_length = accepted_length
        kwargs['max_length'] = max(accepted_length)
        kwargs['verbose_name'] = kwargs.get('verbose_name') or 'GTIN'
        kwargs['validators'] = [GtinValidator(accepted_length=accepted_length)]
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        if value:
            value = ean.compact(value)
        return super().to_python(value)

    def formfield(self, **kwargs):
        return super().formfield(**{
            'form_class': GtinFormField,
            'validators': [GtinValidator(accepted_length=self.accepted_length)],
            **kwargs,
        })



import threading

from bx_py_utils.dict_utils import pluck
from django.db import models

from bx_django_utils.data_types.gtin.model_fields import GtinModelField
from bx_django_utils.models.color_field import ColorModelField
from bx_django_utils.models.timetracking import TimetrackingBaseModel
from bx_django_utils.translation import MultilineTranslationWidget, TranslationField


class TimetrackingTestModel(TimetrackingBaseModel):
    pass


class CreateOrUpdateTestModel(TimetrackingBaseModel):
    name = models.CharField(max_length=64)
    slug = models.SlugField(max_length=64)

    many2one_rel = models.ForeignKey(
        TimetrackingTestModel,
        on_delete=models.CASCADE,
        blank=True, null=True
    )

    blank_field = models.CharField(max_length=64, blank=True)
    null_field = models.CharField(max_length=64, blank=True, null=True)


class GtinFieldTestModel(models.Model):
    default_gtin = GtinModelField()  # accept default length: 12,13 and 14
    all_gtin = GtinModelField(
        blank=True,
        accepted_length=(8, 10, 12, 13, 14)
    )
    ean13 = GtinModelField(
        blank=True, null=True,
        accepted_length=(13,)  # accept one length
    )


class ColorFieldTestModel(models.Model):
    required_color = ColorModelField()
    optional_color = ColorModelField(blank=True, null=True)


class StoreSaveModel(models.Model):
    name = models.CharField(max_length=64)
    _save_calls = threading.local()

    def save(self, **kwargs):
        if not hasattr(self._save_calls, 'saves'):
            self._save_calls.saves = []

        self._save_calls.saves.append(pluck(kwargs, ('arg', 'other_arg')))
        if 'arg' in kwargs:
            del kwargs['arg']
        if 'other_arg' in kwargs:
            del kwargs['other_arg']

        return super().save(**kwargs)


class TranslatedModel(models.Model):
    LANGUAGE_CODES = ['de-de', 'en-us', 'es']

    translated = TranslationField(
        language_codes=LANGUAGE_CODES,
        blank=False,
    )
    translated_multiline = TranslationField(
        language_codes=LANGUAGE_CODES,
        blank=True,
        widget_class=MultilineTranslationWidget,
    )
    not_translated = models.TextField(default='A Default Value')


class RawTranslatedModel(models.Model):
    translated = models.JSONField()
    translated_multiline = models.JSONField(blank=True, default=dict)
    not_translated = models.TextField(default='A Default Value')

    class Meta:
        managed = False
        db_table = TranslatedModel._meta.db_table



from unittest import mock

from django.core.cache import cache
from django.template.defaulttags import CsrfTokenNode
from django.test import TestCase

from bx_django_utils.feature_flags.test_utils import (
    FeatureFlagTestCaseMixin,
    get_feature_flag_cache_info,
    get_feature_flag_db_info,
    get_feature_flag_states,
)
from bx_django_utils.test_utils.html_assertion import HtmlAssertionMixin, assert_html_response_snapshot
from bx_django_utils.test_utils.users import make_test_user
from bx_django_utils_tests.test_app.feature_flags import bar_feature_flag, foo_feature_flag


class FeatureFlagIntegrationTestCase(FeatureFlagTestCaseMixin, HtmlAssertionMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.superuser = make_test_user(is_superuser=True)

    def test_basic(self):
        self.client.force_login(self.superuser)
        response = self.client.get('/admin/')
        self.assert_html_parts(
            response,
            parts=(
                '<a href="/admin/feature_flags/manage/">Manage Feature Flags</a>',
                (
                    '<a href="/admin/feature_flags/feature-flags-values-demo/">'
                    'Feature Flags values DEMO</a>'
                ),
            ),
        )

        with mock.patch.object(CsrfTokenNode, 'render', return_value='MockedCsrfTokenNode'):
            response = self.client.get('/admin/feature_flags/manage/')
        self.assert_html_parts(
            response,
            parts=(
                '<h2>Foo - ENABLED</h2>',
                '<input type="submit" value="Set \'Foo\' to DISABLED">',
                '<h2>Bar - DISABLED</h2>',
                '<input type="submit" value="Set \'Bar\' to ENABLED">',
            ),
        )
        assert_html_response_snapshot(response, query_selector='#content', validate=False)

        response = self.client.get('/admin/feature_flags/feature-flags-values-demo/')
        self.assertRedirects(response, expected_url='/admin/', fetch_redirect_response=False)
        self.assert_messages(
            response,
            expected_messages=[
                'Demo Feature "Foo" state: ENABLED',
                'Demo Feature "Bar" state: DISABLED',
            ],
        )


class PersistantFeatureFlagTestCase(FeatureFlagTestCaseMixin, TestCase):

    def test_peristant(self):
        # Nothing stored, yet:
        self.assertEqual(get_feature_flag_cache_info(), {})
        self.assertEqual(get_feature_flag_db_info(), {})

        # Check initial state of "Foo"-Flag:
        self.assertTrue(foo_feature_flag.is_enabled)  # Foo initial == enabled

        # Check will persistent the current stage of the flag:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-foo': 1})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-foo': 1})

        # Check initial state of "Bar"-Flag:
        self.assertFalse(bar_feature_flag.is_enabled)  # Bar initial == disabled

        # Now both are persistent:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-bar': 0, 'feature-flags-foo': 1})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 0, 'feature-flags-foo': 1})
        # Check get_feature_flag_states():
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': False, 'feature-flags-foo': True})

        # toggle flags:
        foo_feature_flag.disable()
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': False, 'feature-flags-foo': False})
        bar_feature_flag.enable()
        self.assertEqual(get_feature_flag_states(), {'feature-flags-bar': True, 'feature-flags-foo': False})

        # Check state:
        self.assertFalse(foo_feature_flag.is_enabled)
        self.assertTrue(bar_feature_flag.is_enabled)

        # Remove the states from the cache:
        cache.clear()

        # Cache empty, but database holds still the data:
        self.assertEqual(get_feature_flag_cache_info(), {})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})

        # State should be not changed:
        self.assertFalse(foo_feature_flag.is_enabled)
        self.assertTrue(bar_feature_flag.is_enabled)

        # Now it's in the cache:
        self.assertEqual(get_feature_flag_cache_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})
        self.assertEqual(get_feature_flag_db_info(), {'feature-flags-bar': 1, 'feature-flags-foo': 0})



from django.db import models
from django.utils.translation import gettext_lazy as _


class State(models.IntegerChoices):
    """"""  # noqa - don't add in README

    ENABLED = 1, _('enabled')
    DISABLED = 0, _('disabled')



from bx_py_utils.test_utils.snapshot import assert_snapshot
from django import forms
from django.http import HttpResponse
from django.test import TestCase

from bx_django_utils.test_utils.forms import AssertFormFields


class TestUtilsFormsTestCase(TestCase):
    def test_assert_form_fields(self):
        response = self.client.get(path='/admin/login/')

        # Are we on the login page?
        self.assertTemplateUsed(response, 'admin/login.html')

        # Check the form fields:
        assert_form_fields = AssertFormFields(response)
        self.assertEqual(assert_form_fields.get_all_field_names(), {'password', 'username'})
        self.assertGreaterEqual(len(assert_form_fields), 2)

        assert_form_fields.assert_field_names_not_exists(field_names={'internal', 'foo', 'bar'})
        with self.assertRaisesMessage(
            AssertionError, "These fields should not occur, but are present: 'username'"
        ):
            assert_form_fields.assert_field_names_not_exists(
                field_names={'not-exists1', 'username', 'not-exists2'}
            )

        assert_form_fields.assert_field_names_exists(field_names={'password', 'username'})
        with self.assertRaisesMessage(
            AssertionError, "Fields: 'not-exists' are not present in: 'password', 'username'"
        ):
            assert_form_fields.assert_field_names_exists(
                field_names={'password', 'not-exists', 'username'}
            )
        assert_snapshot(got=assert_form_fields.data)

    def test_assert_form_fields_gets_all_field_types(self):
        class TestForm(forms.Form):
            a_input_field = forms.CharField(max_length=123)
            a_hidden_input_field = forms.CharField(widget=forms.HiddenInput)
            a_select_field = forms.ChoiceField(choices=(('foo', 'foo'), ('bar', 'bar')))
            a_textarea = forms.CharField(widget=forms.Textarea)

        content = TestForm().as_p()
        content = f'<div id="content"><form>{content}</form></div>'
        response = HttpResponse(content=f'{content}'.encode())
        assert_form_fields = AssertFormFields(response)
        all_field_names = assert_form_fields.get_all_field_names()
        self.assertEqual(all_field_names, {'a_input_field', 'a_select_field', 'a_textarea'})



from django.core.cache.backends.base import memcache_key_warnings
from django.core.exceptions import ValidationError


def validate_cache_key(cache_key: str) -> None:
    messages = list(memcache_key_warnings(cache_key))
    if messages:
        raise ValidationError(messages)



import uuid
from unittest import TestCase

from bx_py_utils.test_utils.datetime import parse_dt

from bx_django_utils.json_utils import make_json_serializable, to_json


class TestObject:
    __test__ = False

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return f'<TestObject {self.text}>'


def convert_func_example(obj):
    assert isinstance(obj, TestObject)
    return obj.text


class JsonUtilsTestCase(TestCase):
    def test_make_json_serializable(self):
        assert make_json_serializable([
            1, [2, 3], {4: 5, 6: TestObject('foo'), 7: TestObject('bar')}
        ]) == [
            1, [2, 3], {4: 5, 6: '<TestObject foo>', 7: '<TestObject bar>'}
        ]

        assert make_json_serializable(
            [1, [2, 3], {4: 5, 6: TestObject('foo'), 7: TestObject('bar')}],
            convert_func=convert_func_example
        ) == [
            1, [2, 3], {4: 5, 6: 'foo', 7: 'bar'}
        ]

    def test_to_json(self):
        assert to_json([1, 2, 3]) == '[1, 2, 3]'

        assert to_json([
            uuid.UUID('00000000-1111-0000-0000-000000000001'),
            parse_dt('2000-01-01T01:01:01+0100')
        ]) == (
            '["00000000-1111-0000-0000-000000000001", "2000-01-01T01:01:01+01:00"]'
        )

    def test_to_json_convert_func(self):
        assert to_json([TestObject('foo'), TestObject('bar')]) == (
            '["<TestObject foo>", "<TestObject bar>"]'
        )

        assert to_json(
            [TestObject('foo'), TestObject('bar')],
            convert_func=convert_func_example
        ) == '["foo", "bar"]'



# Generated by Django 4.1.7 on 2023-03-10 07:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0003_translatedmodel"),
    ]

    operations = [
        migrations.AddField(
            model_name="translatedmodel",
            name="not_translated",
            field=models.TextField(default="A Default Value"),
        ),
    ]



import inspect
import os
import sys
from importlib import import_module


DEFAULT_EXCLUDED_MODULES = (
    'socketserver',
    'threading',
    'wsgiref',
    'bx_django_utils',
    'django.db',
    'django.core.handlers',
    'django.core.servers',
    'django.utils.decorators',
    'django.utils.deprecation',
    'django.utils.functional',
)


class StackTrace(list):
    pass


class FrameInfo:
    def __init__(self, filename, line, func, code):
        self.filename = filename
        self.line = line
        self.func = func
        self.code = code

    def __str__(self):
        return f'{self.filename} {self.line} {self.func!r} {self.code!r}'

    def __repr__(self):
        return f'<FrameInfo {self}>'


def _exclude(file, excluded_modules):
    excluded_paths = []
    for module in excluded_modules:
        module = import_module(module)
        source_path = inspect.getsourcefile(module)
        if source_path.endswith('__init__.py'):
            source_path = os.path.dirname(source_path)
        excluded_paths.append(os.path.realpath(source_path))

    return any(file.startswith(excluded_path) for excluded_path in excluded_paths)


def iter_frameinfo(start_no=2):
    previous_frame = sys._getframe(start_no)
    while previous_frame:
        file, line, func, code, _ = inspect.getframeinfo(previous_frame, context=1)
        file = os.path.realpath(file)  # make it an absolute path
        if code and (code := code[0]):
            code = code.strip()

        yield (file, line, func, code)

        previous_frame = previous_frame.f_back


def get_stacktrace(tidy=True, exclude_modules=DEFAULT_EXCLUDED_MODULES):
    """
    Returns a StackTrace object, which is a list of FrameInfo objects.
    This construct contains information about the current stacktrace in an easy-to-process format.

    Optionally, you can filter the contained FrameInfo items via the ``tidy`` and ``exclude_modules``.
    Frames from modules in ``exclude_modules`` will we stripped if ``tidy`` is True.
    """
    stacktrace = StackTrace()

    for file, line, func, code in iter_frameinfo():
        if not tidy:
            stacktrace.append(FrameInfo(file, line, func, code))
        if tidy and not _exclude(file, exclude_modules):
            stacktrace.append(FrameInfo(file, line, func, code))

    stacktrace.reverse()
    return stacktrace


class StacktraceAfter:
    """
    Generate a stack trace after a package was visited.
    e.g.:
    After the code path used "django.db", to get the initial code that caused a database query.
    """

    def __init__(self, after_modules):
        self.after_modules = after_modules

    def __call__(self):
        stacktrace = StackTrace()

        before = True  # before we visit self.after_modules modules code
        after = False  # after we left self.after_modules modules code

        for file, line, func, code in iter_frameinfo(start_no=3):
            if before and not _exclude(file, self.after_modules):
                before = False
                continue

            if not after and not _exclude(file, self.after_modules):
                after = True

            if after:
                stacktrace.append(FrameInfo(file, line, func, code))

        stacktrace.reverse()
        return stacktrace



import collections
import re
from collections import Counter
from difflib import unified_diff
from pathlib import Path
from typing import List, Optional, Tuple, Union

from bx_py_utils.test_utils.snapshot import assert_snapshot

from bx_django_utils.dbperf.query_recorder import SQLQueryRecorder
from bx_django_utils.stacktrace import StacktraceAfter


def counter_diff(c1, c2, fromfile=None, tofile=None):
    def pformat(counter):
        lines = []
        for key, value in counter.items():
            lines.append(f'{key}: {value}\n')
        return sorted(lines)

    return ''.join(unified_diff(
        a=pformat(c1),
        b=pformat(c2),
        fromfile=fromfile,
        tofile=tofile
    ))


DEFAULT_AFTER_MODULES = (
    'django.db',
)


class AssertQueries(SQLQueryRecorder):
    """
    Assert executed database queries: Check table names, duplicate/similar Queries.

    Example usage:

        with AssertQueries(query_explain=True) as queries:
            func_that_makes_queries()
        queries.assert_queries(
            query_count=3,
            table_names=['foo','bar'],
            double_tables=True,
            duplicated=True,
            similar=True,
        )

    More example can be found in tests, here:
        bx_django_utils_tests/tests/test_assert_queries.py
    """

    def __init__(
        self,
        databases=None,
        after_modules=None,
        query_explain=False,  # Capture EXPLAIN SQL information?
        max_stacktrace=3,  # How many stack line should be displayed in SQL listing?
    ):
        if after_modules is None:
            after_modules = DEFAULT_AFTER_MODULES

        collect_stacktrace = StacktraceAfter(after_modules=after_modules)
        super().__init__(
            databases=databases, collect_stacktrace=collect_stacktrace, query_explain=query_explain
        )
        assert (
            isinstance(max_stacktrace, int) and max_stacktrace >= 0
        ), f'invalid max_stacktrace: {max_stacktrace!r}'
        self.max_stacktrace = max_stacktrace

    @staticmethod
    def get_table_name(query):
        sql = query['sql']
        if sql.startswith('SAVEPOINT') or sql.startswith('RELEASE SAVEPOINT'):
            # transaction statements
            return

        table_names = re.findall(r'(FROM|INSERT INTO|UPDATE) \"(.+?)\"', sql)
        assert len(table_names) >= 1, f'Error parsing: {sql!r}'
        # Use only the first table name (e.g.: ignore names in inner join)
        table_name = table_names[0][1]
        return table_name

    def count_table_names(self):
        table_name_count = Counter()
        for db, query in self.logger._queries:
            table_name = self.get_table_name(query)
            if table_name:
                table_name_count[table_name] += 1

        return table_name_count

    def get_explains(self):
        """
        Yields DB table name and captured SQL explain information.
        """
        assert self.query_explain, 'Explain way not captured!'

        for db, query in self.logger._queries:
            table_name = self.get_table_name(query)
            explain_str = '\n'.join(query['explain'])

            yield table_name, explain_str

    def get_explain_dict(self):
        """
        Generate a dict with database table name and SQL explain entry.
        """
        result = collections.defaultdict(list)
        for table_name, explain_str in self.get_explains():
            result[table_name].append(explain_str)
        return dict(result)

    @property
    def query_info(self) -> str:
        """
        :return: Human readable information about the executed SQL queries
        """
        parts = []
        for i, (db, query) in enumerate(self.logger._queries, start=1):
            parts.append(f'{i:>3}. {query["sql"]}')

            if self.max_stacktrace:
                stacktrace = query['stacktrace']
                stack_count = len(stacktrace)
                if stack_count > self.max_stacktrace:
                    parts.append('    ...')
                else:
                    parts.append('\n')

                max_lines_slice = -self.max_stacktrace
                stacktrace = stacktrace[max_lines_slice:]

                start_no = stack_count - self.max_stacktrace + 1
                for frame_no, frameinfo in enumerate(stacktrace, start=start_no):
                    parts.append(
                        f'    {frame_no:>2}.'
                        f'{frameinfo.filename} {frameinfo.line} {frameinfo.func}():'
                    )
                    parts.append(f'        {frameinfo.code!r}')
                parts.append('\n')

            if self.query_explain:
                for explain_line in query['explain']:
                    parts.append(f'    {explain_line}')
                parts.append('')

        return '\n'.join(parts)

    def build_error_message(self, msg):
        return (
            f'{msg}\n'
            f'Captured queries were:\n'
            f'{self.query_info}'
        )

    def assert_table_names(self, *expected_table_names):
        """
        Check tables are used.
        """
        expected_table_names = set(expected_table_names)

        table_name_count = self.count_table_names()
        used_tables = set(table_name_count.elements())

        if expected_table_names != used_tables:
            diff = ''.join(unified_diff(
                sorted(f'{n}\n' for n in expected_table_names),
                sorted(f'{n}\n' for n in used_tables),
                fromfile='expected table names',
                tofile='used table names'
            ))
            raise AssertionError(self.build_error_message(f'Table names does not match:\n{diff}'))

    def assert_table_counts(self, table_counts: Union[Counter, dict],
                            exclude: Optional[Tuple[str]] = None):
        if not isinstance(table_counts, Counter):
            table_counts = Counter(table_counts)
        table_name_count = self.count_table_names()
        if exclude:
            for k in exclude:
                assert k not in table_counts, f'Excluded key {k!r} is listed in table_counts'
                if k in table_name_count:
                    del table_name_count[k]
        if table_counts != table_name_count:
            diff = counter_diff(
                c1=table_counts,
                c2=table_name_count,
                fromfile='expected table counts',
                tofile='current table counts'
            )
            raise AssertionError(self.build_error_message(f'Table count error:\n{diff}'))

    def snapshot_table_counts(self, *, exclude: Optional[Tuple[str]] = None, **kwargs):
        table_name_count = self.count_table_names()
        if exclude:
            for k in exclude:
                if k in table_name_count:
                    del table_name_count[k]
        assert_snapshot(got=table_name_count, self_file_path=Path(__file__), **kwargs)

    def assert_not_double_tables(self):
        """
        Check if tables are used more than one time.
        """
        table_name_count = self.count_table_names()

        double_names = {
            table_name
            for table_name, count in table_name_count.items()
            if count > 1
        }
        assert not double_names, self.build_error_message(
            f'These tables was double used:\n{double_names}'
        )

    def assert_duplicated_queries(self, duplicated=True, similar=True):
        """
        Check similar and/or duplicated queries
        """
        assert duplicated or similar

        results = self.logger.dump(aggregate_queries=True)
        num_queries_duplicated = results['num_queries_duplicated']
        num_queries_similar = results['num_queries_similar']

        if duplicated and similar:
            assert num_queries_duplicated == 0 and num_queries_similar == 0, \
                self.build_error_message(
                    f'There are {num_queries_duplicated} duplicated'
                    f' and {num_queries_similar} similar queries.'
                )
        elif duplicated:
            assert num_queries_duplicated == 0, self.build_error_message(
                f'There are {num_queries_duplicated} duplicated queries.'
            )
        elif similar:
            assert num_queries_similar == 0, self.build_error_message(
                f'There are {num_queries_similar} similar queries.'
            )

    def assert_query_count(self, num):
        """
        Check the total executed database query count.
        Similar to: self.assertNumQueries(num=123)
        """
        captured_queries = self.logger._queries
        queries_count = len(captured_queries)
        if num != queries_count:
            raise AssertionError(
                self.build_error_message(f'{queries_count} queries executed, {num} expected.')
            )

    def assert_queries(
        self,
        table_counts: Optional[Union[Counter, dict]] = None,
        double_tables: Optional[bool] = True,
        table_names: Optional[List[str]] = None,
        query_count: Optional[int] = None,
        duplicated: Optional[bool] = True,
        similar: Optional[bool] = True,
    ):
        """
        Shortcut to assert everything
        """
        if table_counts is not None:
            self.assert_table_counts(table_counts=table_counts)

        if double_tables:
            self.assert_not_double_tables()

        if table_names is not None:
            self.assert_table_names(*table_names)

        if query_count is not None:
            self.assert_query_count(num=query_count)

        if duplicated or similar:
            self.assert_duplicated_queries(duplicated=duplicated, similar=similar)



from typing import Iterator, Type, Union

from django.http import HttpRequest
from django.urls import path, reverse
from django.views import View

from bx_django_utils.admin_extra_views.base_view import AdminExtraViewMixin
from bx_django_utils.admin_extra_views.datatypes import AdminExtraMeta, PseudoApp


class AdminExtraViewRegistry:
    """
    Hold all information about all admin extra views to expand urls and admin app list.
    """

    def __init__(self):
        self.pseudo_apps = set()

    def add_view(
        self, pseudo_app: PseudoApp, view_class: Type[Union[AdminExtraViewMixin, View]]
    ) -> None:
        """
        Collect all admin extra views. Called by our @register_admin_view() decorator.
        """
        if not issubclass(view_class, AdminExtraViewMixin):
            raise ValueError('Wrapped class must subclass AdminExtraViewMixin.')

        if not isinstance(view_class.meta, AdminExtraMeta):
            raise ValueError(f'View {view_class} must have a "AdminExtraMeta" instance.')

        view_class.meta.setup_app(pseudo_app)

        pseudo_app.views.append(view_class)
        if pseudo_app not in self.pseudo_apps:
            self.pseudo_apps.add(pseudo_app)

    def get_urls(self) -> list:
        """
        :return: all url patterns for all admin extra views.

        Should be used in urls.py beside the "admin.site.urls" to add all extra views.
        """
        urls = []

        for pseudo_app in self.pseudo_apps:
            for view_class in pseudo_app.views:
                url_pattern = path(
                    view_class.meta.url,
                    view_class.as_view(),
                    name=view_class.meta.url_name,
                )
                urls.append(url_pattern)

        return urls

    def get_app_list(self, request: HttpRequest) -> list:
        """
        :return: The extra views app list, filtered by all conditions

        Should be used in AdminSite.get_app_list() to add all extra views.
        """
        app_list = []
        for pseudo_app in sorted(self.pseudo_apps, key=lambda x: x.meta.name):
            if not all(cond(request) for cond in pseudo_app.meta.conditions):
                # Don't add the view if the user can't use them
                # Important: This will not deny the access!
                # conditions will be checked in AdminExtraViewMixin.dispatch(), too!
                continue

            models = []
            for view_class in sorted(pseudo_app.views, key=lambda x: x.meta.name):
                if all(cond(request) for cond in view_class.meta.conditions):
                    models.append(
                        {
                            'name': view_class.meta.name,
                            'app_label': view_class.meta.app_label,
                            'admin_url': reverse(view_class.meta.url_name),
                            'view_only': True,
                        }
                    )
            if models:
                app_list.append(
                    {
                        'name': pseudo_app.meta.name,
                        'app_label': pseudo_app.meta.app_label,
                        'models': models,
                    }
                )

        return app_list

    def __iter__(self) -> Iterator[Type[Union[AdminExtraViewMixin, View]]]:
        """
        Iterate sorted over all registered admin extra view classes.
        """
        for pseudo_app in sorted(self.pseudo_apps, key=lambda x: x.meta.name):
            yield from sorted(pseudo_app.views, key=lambda x: x.meta.name)


extra_view_registry = AdminExtraViewRegistry()


def register_admin_view(*, pseudo_app):
    """
    Decorator to add a normal view as pseudo App/Model to the admin.
    """

    def _model_admin_wrapper(view_class):
        if not isinstance(pseudo_app, PseudoApp):
            raise ValueError('Keyword argument "pseudo_app" must be a PseudoApp instance.')

        extra_view_registry.add_view(pseudo_app, view_class)

        return view_class

    return _model_admin_wrapper



from collections import defaultdict
from pprint import saferepr
from typing import Callable, List, Optional

from django.db import connections

from bx_django_utils.dbperf.cursor import RecordingCursorWrapper


class Logger:
    def __init__(self):
        self._queries = []  # list of (dbname, metrics) tuples for each query that was run
        self._databases = {}  # short summary
        self._sql_time = 0  # total execution time of all queries in milliseconds
        self._num_queries = 0  # total count of queries executed

    def record(self, alias, **kwargs):
        self._queries.append((alias, kwargs))

        if alias not in self._databases:
            self._databases[alias] = {
                "time_spent": kwargs["duration"],
                "num_queries": 1,
            }
        else:
            self._databases[alias]["time_spent"] += kwargs["duration"]
            self._databases[alias]["num_queries"] += 1

        self._sql_time += kwargs['duration']
        self._num_queries += 1

    @staticmethod
    def _similar_key(query):
        return query['raw_sql']

    @staticmethod
    def _duplicate_key(query):
        raw_params = () if query['raw_params'] is None else tuple(query['raw_params'])
        # saferepr() avoids problems because of unhashable types
        # (e.g. lists) when used as dictionary keys.
        return query['raw_sql'], saferepr(raw_params)

    def _aggregate(self, results):
        aliases = set()
        # todo: defaultdicts handle very awkwardly (e.g. with Django Templates),
        #  is collections.Counter a good replacement?
        queries_similar = defaultdict(lambda: defaultdict(int))
        queries_duplicated = defaultdict(lambda: defaultdict(int))
        num_queries_similar = 0
        num_queries_duplicated = 0

        # put every query into similar/duplicated dicts to aggregate their execution count
        for alias, query in self._queries:
            aliases.add(alias)
            queries_similar[alias][self._similar_key(query)] += 1
            queries_duplicated[alias][self._duplicate_key(query)] += 1

        # now that the aggregation is done we must strip all queries in the
        # aggregation dicts that were executed only once.
        for agg_dict in (queries_similar, queries_duplicated):
            for db, queries in agg_dict.items():
                del_keys = []
                for query, count in queries.items():
                    if count == 1:
                        del_keys.append(query)
                # map(lambda key: queries.pop(key), del_keys)
                for key in del_keys:
                    del queries[key]

        # for convenience, make a total for each aggregation across all databases and queries
        for db in queries_similar:
            num_queries_similar += len(queries_similar[db])
        for db in queries_duplicated:
            num_queries_duplicated += len(queries_duplicated[db])

        results['queries_similar'] = queries_similar
        results['queries_duplicated'] = queries_duplicated
        results['num_queries_similar'] = num_queries_similar
        results['num_queries_duplicated'] = num_queries_duplicated

    def dump(self, aggregate_queries=True):
        results = {
            'queries': self._queries,
            'databases': self._databases,
            'sql_time': self._sql_time,
            'num_queries': self._num_queries,
        }

        if aggregate_queries:
            self._aggregate(results)

        return results


class SQLQueryRecorder:
    """
    A context manager that allows recording SQL queries executed during its lifetime.
    Results of the recording can be read from the recorder in various formats after
    recording has stopped.

    In environments where not all databases in settings.DATABASES are available (e.g. unittests)
    you should pass the databases you want to observe to the constructor.

    Example usage:

        with SQLQueryRecorder(query_explain=True) as rec:
            func_that_makes_queries()
        print(rec.results(aggregate_results=True))

    """
    running = None

    def __init__(
        self,
        databases: Optional[List[str]] = None,
        collect_stacktrace: Optional[Callable] = None,
        query_explain: bool = False,  # Capture EXPLAIN SQL information?
    ):
        self.logger = Logger()
        self.query_explain = query_explain

        if databases:
            self.databases = [db for db in connections.all() if db.alias in databases]
        else:
            self.databases = connections.all()

        self.collect_stacktrace = collect_stacktrace

    def __enter__(self):
        for connection in self.databases:
            assert not hasattr(connection, '_recording_cursor'), \
                'the connection\'s cursor has not been unwrapped properly in a previous run'

            # wrap Django's default cursors in RecordingCursorWrapper
            connection._recording_cursor = connection.cursor
            connection._recording_chunked_cursor = connection.chunked_cursor

            def cursor():
                return RecordingCursorWrapper(
                    connection._recording_cursor(),
                    connection,
                    self.logger,
                    collect_stacktrace=self.collect_stacktrace,
                    query_explain=self.query_explain,
                )

            def chunked_cursor():
                return RecordingCursorWrapper(
                    connection._recording_chunked_cursor(),
                    connection,
                    self.logger,
                    collect_stacktrace=self.collect_stacktrace,
                    query_explain=self.query_explain,
                )

            connection.cursor = cursor
            connection.chunked_cursor = chunked_cursor

        self.running = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for connection in self.databases:
            assert hasattr(connection, '_recording_cursor'), \
                'the connection has already been unwrapped, this should not have happened'

            # undo the cursor wrapping so the connection is 'clean' again
            del connection._recording_cursor
            del connection.cursor
            del connection.chunked_cursor

        self.running = False

    def results(self, **dump_kwargs):
        assert not self.running, 'results can only be accessed after recording has stopped'
        return self.logger.dump(**dump_kwargs)



import json
import pathlib
import tempfile
from collections import Counter

from bx_py_utils.environ import OverrideEnviron
from django.contrib.auth.models import Group, Permission
from django.db import connection, transaction
from django.test import TestCase

from bx_django_utils.test_utils.assert_queries import AssertQueries


def make_database_queries(count=1):
    for i in range(count):
        Permission.objects.all().first()


def make_database_queries2(count=1):
    make_database_queries(count)


class AssertQueriesTestCase(TestCase):

    def get_instance(self):
        with AssertQueries() as queries:
            Permission.objects.all().first()
        return queries

    def test_assert_queries(self):
        queries = self.get_instance()
        queries.assert_queries(
            query_count=1,
            table_names=['auth_permission'],
            double_tables=True,
            duplicated=True,
            similar=True,
        )

    def test_assert_table_names(self):
        queries = self.get_instance()
        queries.assert_table_names('auth_permission')

        with self.assertRaises(AssertionError) as err:
            queries.assert_table_names('foo', 'bar')
        msg = str(err.exception)
        assert 'Table names does not match:' in msg
        assert '-bar\n' in msg
        assert '-foo\n' in msg
        assert '+auth_permission\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg

    def test_assert_table_counts(self):
        queries = self.get_instance()
        queries.assert_table_counts(Counter(auth_permission=1))
        queries.assert_table_counts({'auth_permission': 1})

        with self.assertRaises(AssertionError) as err:
            queries.assert_table_counts(Counter(auth_permission=2, foo=1, bar=3))
        msg = str(err.exception)
        assert 'Table count error:\n' in msg
        assert '-auth_permission: 2\n' in msg
        assert '-bar: 3\n' in msg
        assert '-foo: 1\n' in msg
        assert '+auth_permission: 1\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg

    def test_assert_table_counts_exclude(self):
        with AssertQueries() as queries:
            Permission.objects.all().first()
            Permission.objects.all().first()
            Group.objects.all().first()

        queries.assert_table_counts(
            {'auth_permission': 2}, exclude=('auth_group', 'not_mentioned')
        )

        # Incorrect counts / added tables
        with self.assertRaises(AssertionError) as err:
            queries.assert_table_counts({'auth_permission': 3, 'foo': 1}, exclude=('auth_group',))
        msg = str(err.exception)
        assert 'Table count error:\n' in msg
        assert '-auth_permission: 3\n' in msg
        assert '-foo: 1\n' in msg
        assert '+auth_permission: 2\n' in msg
        assert 'Captured queries were:\n' in msg

        # Missing counts
        with self.assertRaises(AssertionError) as err:
            queries.assert_table_counts({'auth_group': 1}, exclude=('foobar',))
        assert 'Table count error:\n' in msg
        assert '+auth_permission: 2\n' in msg

        # Excluding a table that is specified
        with self.assertRaises(AssertionError) as err:
            queries.assert_table_counts(
                {'auth_permission': 3, 'auth_group': 1}, exclude=('auth_permission',))
        msg = str(err.exception)
        assert msg == 'Excluded key \'auth_permission\' is listed in table_counts'

    def test_snapshot_table_counts(self):
        with AssertQueries() as queries:
            Permission.objects.all().first()
            Permission.objects.all().first()
            Group.objects.all().first()

        with AssertQueries() as second_queries:
            Permission.objects.all().first()
            Permission.objects.all().first()
            Permission.objects.all().first()
            Group.objects.all().first()

        with tempfile.TemporaryDirectory() as tmp_dir, OverrideEnviron(
            RAISE_SNAPSHOT_ERRORS='1'  # Maybe it's disabled in this test run!
        ), self.assertRaises(FileNotFoundError):
            queries.snapshot_table_counts(
                exclude=('auth_group',),
                root_dir=tmp_dir,
                snapshot_name='test',
            )

            queries.snapshot_table_counts(
                exclude=('auth_group',), root_dir=tmp_dir, snapshot_name='test'
            )

            snapshot_file = pathlib.Path(tmp_dir) / 'test.snapshot.json'
            snapshot_data = json.loads(snapshot_file.read_text())
            self.assertEqual(
                snapshot_data,
                {
                    'auth_permission': 2,
                },
            )

            with self.assertRaises(AssertionError) as error_capture:
                second_queries.snapshot_table_counts(
                    exclude=('auth_group',), root_dir=tmp_dir, snapshot_name='test'
                )

            message = str(error_capture.exception)
            self.assertIn('"auth_permission": 2', message)
            self.assertIn('"auth_permission": 3', message)

        # Use actual test files
        queries.snapshot_table_counts(exclude=('auth_group',))
        second_queries.snapshot_table_counts()

    def test_assert_not_double_tables(self):
        queries = self.get_instance()
        queries.assert_not_double_tables()

        with self.assertRaises(AssertionError) as err:
            with AssertQueries() as queries:
                Permission.objects.all().first()
                Permission.objects.all().first()
            queries.assert_not_double_tables()
        msg = str(err.exception)
        assert 'These tables was double used:\n' in msg
        assert "{'auth_permission'}\n" in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg
        assert '2. SELECT "auth_permission"' in msg

    def test_assert_duplicated_queries(self):
        queries = self.get_instance()
        queries.assert_duplicated_queries(duplicated=True, similar=True)

        # duplicated + similar

        with self.assertRaises(AssertionError) as err, \
                AssertQueries() as queries:
            Permission.objects.all().first()  # Query 1
            Permission.objects.all().first()  # Query 2
            queries.assert_duplicated_queries(duplicated=True, similar=True)
        msg = str(err.exception)
        assert 'There are 1 duplicated and 1 similar queries.\n' in msg
        assert 'Captured queries were:\n' in msg

        assert '1. SELECT "auth_permission"' in msg
        assert 'bx_django_utils_tests/tests/test_assert_queries.py' in msg
        assert ' test_assert_duplicated_queries():\n' in msg
        assert "Permission.objects.all().first()  # Query 1" in msg

        assert '2. SELECT "auth_permission"' in msg
        assert "Permission.objects.all().first()  # Query 2" in msg

        # no duplicated, but similar

        with self.assertRaises(AssertionError) as err, \
                AssertQueries() as queries:
            Permission.objects.exclude(name='query 3').first()
            Permission.objects.exclude(name='query 4').first()
            queries.assert_duplicated_queries(duplicated=True, similar=True)
        msg = str(err.exception)
        assert 'There are 0 duplicated and 1 similar queries.\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg
        assert "Permission.objects.exclude(name='query 3').first()" in msg
        assert '2. SELECT "auth_permission"' in msg
        assert "Permission.objects.exclude(name='query 4').first()" in msg

        # only duplicated

        with self.assertRaises(AssertionError) as err, \
                AssertQueries() as queries:
            Permission.objects.all().first()  # Query 5
            Permission.objects.all().first()  # Query 6
            queries.assert_duplicated_queries(duplicated=True, similar=False)
        msg = str(err.exception)
        assert 'There are 1 duplicated queries.\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg
        assert "Permission.objects.all().first()  # Query 5" in msg
        assert '2. SELECT "auth_permission"' in msg
        assert "Permission.objects.all().first()  # Query 6" in msg

        # only similar

        with self.assertRaises(AssertionError) as err, \
                AssertQueries() as queries:
            Permission.objects.exclude(name='query 7').first()
            Permission.objects.exclude(name='query 8').first()
            queries.assert_duplicated_queries(duplicated=False, similar=True)
        msg = str(err.exception)
        assert 'There are 1 similar queries.\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg
        assert "Permission.objects.exclude(name='query 7').first()" in msg
        assert '2. SELECT "auth_permission"' in msg
        assert "Permission.objects.exclude(name='query 8').first()" in msg

    def test_assert_query_count(self):
        queries = self.get_instance()
        queries.assert_query_count(num=1)

        with self.assertRaises(AssertionError) as err:
            queries.assert_query_count(num=2)
        msg = str(err.exception)
        assert '1 queries executed, 2 expected.\n' in msg
        assert 'Captured queries were:\n' in msg
        assert '1. SELECT "auth_permission"' in msg

    def test_assert_queries_variants(self):

        # INSERT INTO "auth_group" ("name") VALUES (%s)

        with AssertQueries() as queries:
            Group.objects.create(name='foo')
        queries.assert_queries(
            table_counts=Counter({
                'auth_group': 1
            }),
            double_tables=True,
            table_names=['auth_group'],
            duplicated=True,
            similar=True,
        )

        # SELECT "auth_group"."id", "auth_group"."name" FROM...

        with AssertQueries() as queries:
            group = Group.objects.first()
        queries.assert_queries(
            table_counts=Counter({
                'auth_group': 1
            }),
            double_tables=True,
            table_names=['auth_group'],
            duplicated=True,
            similar=True,
        )

        # UPDATE "auth_group" SET "name" = %s WHERE...

        with AssertQueries() as queries:
            group.name = 'foo bar'
            group.save(update_fields=['name'])
        queries.assert_queries(
            table_counts=Counter({
                'auth_group': 1
            }),
            double_tables=True,
            table_names=['auth_group'],
            duplicated=True,
            similar=True,
        )

        # SAVEPOINT...RELEASE SAVEPOINT

        with AssertQueries() as queries:
            with transaction.atomic():
                group.name = 'foobar'
                group.save()

        queries.assert_queries(
            table_counts=Counter({
                'auth_group': 1
            }),
            double_tables=True,
            table_names=['auth_group'],
            duplicated=True,
            similar=True,
        )

        # INNER JOIN tables will be ignored:

        with AssertQueries() as queries:
            Group.objects.filter(
                permissions__in=Permission.objects.filter(content_type__app_label='not-exist')
            ).count()

        queries.assert_queries(
            table_counts=Counter({
                'auth_group': 1
            }),
            double_tables=True,
            table_names=['auth_group'],
            duplicated=True,
            similar=True,
        )

    def test_query_explain(self):
        with AssertQueries(query_explain=True) as queries:
            Permission.objects.all().first()

        results = list(queries.get_explains())
        self.assertEqual(len(results), 1)
        table_name, explain_str = results[0]

        self.assertEqual(table_name, 'auth_permission')
        self.assertIsInstance(explain_str, str)
        if connection.vendor == 'sqlite':
            # Depends on SQLite version/implementation!
            self.assertIn(' SCAN ', explain_str)
            self.assertIn(' USING INDEX ', explain_str)
            self.assertIn(' SEARCH ', explain_str)

    def test_query_info(self):
        with AssertQueries(query_explain=True) as queries:
            Permission.objects.all().first()
            Group.objects.all().first()

        query_info = queries.query_info
        self.assertIsInstance(query_info, str)
        self.assertIn('1. SELECT ', query_info)
        self.assertIn('2. SELECT ', query_info)
        self.assertIn('bx_django_utils_tests/tests/test_assert_queries.py ', query_info)
        self.assertIn(' test_query_info():\n', query_info)
        if connection.vendor == 'sqlite':
            # Depends on SQLite version/implementation!
            self.assertIn(' SCAN ', query_info)
            self.assertIn(' USING INDEX ', query_info)
            self.assertIn(' SEARCH ', query_info)

    def test_query_explain_not_captured(self):
        with AssertQueries() as queries:
            Permission.objects.all().first()

        with self.assertRaisesMessage(AssertionError, 'Explain way not captured!'):
            list(queries.get_explains())

    def test_query_info_max_stacktrace(self):
        with AssertQueries(max_stacktrace=2) as queries:
            make_database_queries2(count=1)

        query_info = queries.query_info
        self.assertIsInstance(query_info, str)
        self.assertIn('bx_django_utils_tests/tests/test_assert_queries.py ', query_info)
        self.assertIn('make_database_queries2():\n', query_info)  # <<< last two frames present
        self.assertIn('make_database_queries():\n', query_info)  # last frame

        with AssertQueries(max_stacktrace=1) as queries:
            make_database_queries2(count=1)

        query_info = queries.query_info
        self.assertIsInstance(query_info, str)
        self.assertIn('bx_django_utils_tests/tests/test_assert_queries.py ', query_info)
        self.assertNotIn('make_database_queries2():', query_info)  # <<< only last -> not present
        self.assertIn('make_database_queries():\n', query_info)  # last frame
